/*
 * mcolors.c: device-independent color routines.
 */

#include "memmon.h"

hidden novalue setcolor Params((char *label, char *value));

/*
 *  Color labels and codes.
 */
static struct centry {
   char *label;			/* label, for legend or recoloring */
   char code;			/* block code */
   };

#define Color(sym, code, lab) { lab, code },

static struct centry cdata[NColors] = {
#include "msymbols.h"
   };

/*
 *  Default color values.
 */
struct blkcolor {
   char *label;			/* label */
   char value[4];		/* value, as a 3-char string */
   };

#define ColorDefault(label, spec) { label, spec },

static struct blkcolor defcolors [] = {
#include "mcolors.h"
   { 0, "" }			/* terminator */
   };

/* 
 * initcolors() - initialize color system.
 */
novalue initcolors()
   {
   int i;
   struct blkcolor *v;

   /* initialize char-to-color mapping */
   for (i = 0; i < NColors; i++)
      if (cdata[i].code != 0)
         blkcolor[cdata[i].code] = i;	/* set color for character */

   /* set color defaults */
   for (v = defcolors; v->label; v++)
      setcolor(v->label, v->value);
   }

/* 
 * readcolors(filename) - read a color specification file.
 */

#define ColrArgs 2	/* number of args on a color spec line */

novalue readcolors(filename)
char *filename;
   {
   FILE *f;
   int n;
   char line[LineSize];
   char *p;
   char *w[ColrArgs];

   if (strcmp(filename, "-") == 0)
      f = stdin;
   else
      f = fopen(filename, "r");
   if (!f)
      pexit(filename);
   while (fgets(line, LineSize, f)) {
      p = trim(line, '#');
      n = chop(p, w, ColrArgs, 0);
      if (n > 0)
         setcolor(w[0], w[1]);
      }
   if (strcmp(filename, "-") == 0)
      clearerr(f);
   else
      fclose(f);
   }

/*
 * setcolor(label, value) - set a color in the color map.
 */
static novalue setcolor(label, value)
char *label;
char *value;
   {
   int i, r, g, b;
   static int map8[] = { 0, 36, 73, 109, 146, 182, 219, 255 };

   for (i = 0; i < NColors; i++)
      if (strcmp(label, cdata[i].label) == 0)
         break;
   if (i >= NColors) {
      fprintf(stderr, "%s: unrecognized color label: %s\n", progname, label);
      return;
      }
   if (strlen(value) != 3 || sscanf(value, "%1o%1o%1o", &r, &g, &b) != 3) {
      fprintf(stderr, "%s: bad color value: %s %s\n", progname, label, value);
      return;
      }
   cmap[i].red = map8[r];
   cmap[i].green = map8[g];
   cmap[i].blue = map8[b];
   }

/*
 * setmap(m, c) - set Marked or Unmarked colors to c, according to m, and set
 *  others back to normal.  Load the resulting color map into the device.
 */
novalue setmap(m, c)
int m, c;
   {
   int i, n;
   
   n = Marked + Unmarked - m;		/* n is opposite of m (Marked or Un) */
   for (i = 0; i < NColors; i++)  {	/* alter map buffer */
      cmap[n + i] = cmap[i];
      cmap[m + i] = cmap[c];
      }
   devmap();				/* load map into device */
   }

/*
 * legend() - display legend
 */
novalue legend()
   {
   struct centry *e;
   int col, n, lum, fg;

   col = 0;
   for (e = cdata; e->code; e++) {
      if (e->code != SpaceCode) {
         n = e - cdata;
         lum = 30 * cmap[n].red + 59 * cmap[n].green + 11 * cmap[n].blue;
         fg = (lum > (100 * 128)) ? C_Black : C_White;
         mtext(e->label, 1, col, fg, e - cdata);
         }
      col += strlen(e->label) + 1;
      }
   }
