/*
 * mmps.c: batch mode PostScript translator.
 */

#include "memmon.h"

hidden novalue pscolor Params((int c));

#define PageHeight 792
#define PageWidth 612
#define SideMargin 72
#define TopMargin 72

#define Font "Helvetica"
#define CharHeight (3 * textrow / 4)	/* height of a text char */
#define CharBase   (textrow / 4)	/* offset to baseline within text box */

static int llx, lly, urx, ury;		/* bounding box */
static float psscale;			/* output scaling */
static int lastpscolor = -1;		/* last color command output */

/*
 * devsetup() - set globals to device-dependent values.
 */
novalue devsetup()
   {
   granularity = 4;
   width = PageWidth - 2 * SideMargin;
   height = PageHeight - 2 * TopMargin;
   textrow = 11;
   textsep = 2;
   memrow = 20;
   batchmode = 1;
   }

/*
 * devinit() - initialize output file.
 */
novalue devinit()
   {
   char time_buf[26];

   psscale = 1.0;
   if (width > PageWidth - 2 * SideMargin)
      psscale = (PageWidth - 2 * SideMargin) / (float)width;
   if (psscale * height > PageHeight - 2 * TopMargin)
      psscale = (PageHeight - 2 * TopMargin) / (psscale * height);
   llx = (PageWidth - psscale * width) / 2;
   lly = PageHeight - TopMargin - psscale * height;
   urx = llx + psscale * width + 0.99;
   ury = lly + psscale * height + 0.99;

   width -= 2;				/* for border */
   height -= 2;				/* for border */

   getctime(time_buf);
   printf("%%!PS-Adobe-2.0 EPSF-1.2\n");
   if (title)
      printf("%%%%Title: %s\n", title);
   printf("%%%%Creator: mmps\n");
   printf("%%%%CreationDate: %s", time_buf);
   printf("%%%%DocumentFonts: %s\n", Font);
   printf("%%%%BoundingBox: %d %d %d %d\n", llx, lly, urx, ury);
   printf("%%%%EndComments\n");
   printf
      ("/B { moveto 0 H rlineto 0 rlineto 0 H neg rlineto fill } bind def\n");
   printf("/C { setrgbcolor } bind def\n");
   printf("/S { moveto show } bind def\n");
   printf("%%%%EndProlog\n");
   }

/*
 * batbegin() - prepare to write an image.
 */
novalue batbegin()
   {
   static int pageno = 0;

   ++pageno;
   lastpscolor = -1;
   printf("%%%%Page: %d %d\n", pageno, pageno);
   printf("save\n");
   pscolor(C_Background);
   printf("%d %d moveto %d %d lineto %d %d lineto %d %d lineto fill\n",
      llx, lly, llx, ury, urx, ury, urx, lly);
   printf("%d %d translate ", llx, lly);
   printf("%f %f scale ", psscale, psscale);
   printf("1 1 translate\n");		/* for border */
   printf("1 setlinewidth\n");
   if (textrow > 0) {
      printf("/%s findfont [%.3f 0 0 %.3f 0 0] makefont setfont\n", Font,
         1.9 * (float)width / (float)TextLength, (float)CharHeight);
      printf("/H %d def\n", textrow - 1);
      }
   }

/*
 * battext(s, row, col, fg, bg) - output text string.
 */
novalue battext(s, row, col, fg, bg)
char *s;
int row, col, fg, bg;
   {
   float charwidth;
   int x, xx, y;

   charwidth = (float)width / (float)TextLength;
   x = charwidth * col;
   xx = charwidth * (col + strlen(s) + 1);
   y = height - (row + 1) * textrow + 1;
   pscolor(bg);
   printf("%d %d %d B ", xx - x - 1, x, y);
   pscolor(fg);
   pstext(s);
   printf(" %d %d S\n", (int)(x + charwidth / 2), y + CharBase);
   }

/*
 * batmem(mbuf) - output memory image.
 */
novalue batmem(mbuf)
unsigned char mbuf[];
   {
   int b, c, d, n, x, y;
   unsigned char *p, *endrow, *endmem;

   if (memrow > 1)
      printf("/H %d def\n", memrow - 1);	/* define row height */
   else
      printf("/H 1 def\n");

   /*
    * identify contiguous blocks of color
    * (assumes we can skip the separator pixel between two blocks)
    */
   p = mbuf;
   endmem = mbuf + mempixels;
   y = memheight;
   while (p < endmem)  {
      endrow = p + width;
      if (endrow > endmem)
         endrow = endmem;
      x = -1;
      y -= memrow;
      while (p < endrow) {
         c = *p++;
         x++;
         if (c == C_Bsep)
            continue;
         n = 1;
         while (p < endrow && (d = *p++) == c)
            n++;
         pscolor(c);			/* set color */
         b = c % NColors;		/* color sans Marked/Unmarked flag */
         if (p == endrow || b == C_String || b == C_Ssep)
            printf("%d %d %d B\n", n, x, y);	/* write string block exactly */
         else
            printf("%d %d %d B\n", n-1, x, y);	/* otherwise, emphasize gap */
         x += n;
         if (d != C_Bsep && p < endrow)
            p--, x--;
         }
      }

   printf("restore showpage\n");
   }

/*
 * batterm() - terminate entire run.
 */
novalue batterm()
   {
   printf("%%%%Trailer\n");
   }

/*
 * pscolor(c) - write a setcolor command for color c,
 *  if it's not the same as the last one.
 */
static novalue pscolor(c)
int c;
   {
   if (lastpscolor == c)
      return;
   printf("%.2f %.2f %.2f C ",
      cmap[c].red / 255.0, cmap[c].green / 255.0, cmap[c].blue / 255.0);
   lastpscolor = c;
   }
