/*
 * This files contains routines for getting the "next" token.
 */
#include "preproc.h"

/*
 * next_tok - get the next raw token. No macros have been expanded. Only
 *  #line directives are been processed.
 */
struct token *next_tok()
   {
   struct token *t, *t1;
   struct tok_lst *tlst;
   struct paste_lsts *plsts;
   struct char_src *cs;
   union src_ref ref;
   struct str_buf *sbuf;
   char *s;
   char *fname;
   int n, i;

   if (src_stack->flag == DummySrc)
      return NULL;

   /*
    * See if a directive pushed back any tokens.
    */
   if (src_stack->ntoks > 0)
      return src_stack->toks[--src_stack->ntoks];

   switch (src_stack->flag) {
      case CharSrc:
         t = tokenize();
         if (t != NULL && src_stack->u.cs->f != NULL)
            t->flag |= LineChk;
         if (t != NULL && t->tok_id == PpLine) {
            /*
             * #line directives must be processed here so they are not
             *  put in macros.
             */
            cs = src_stack->u.cs;
            t1 = NULL;
            advance_tok(&t1);
            if (t1->tok_id != PpNumber)
               errt1(t1, "#line requires an integer argument");
            n = 0;
            for (s = t1->image; *s != '\0'; ++s) {
               if (*s >= '0' && *s <= '9')
                  n = 10 * n + (*s - '0');
               else
                  errt1(t1, "#line requires an integer argument");
               }
            advance_tok(&t1);
            fname = NULL;
            if (t1->tok_id == StrLit) {
               sbuf = get_sbuf();
               for (s = t1->image; *s != '\0'; ++s) {
                  if (s[0] == '\\' && (s[1] == '\\' || s[1] == '"'))
                     ++s;
                  AppChar(*sbuf, *s);
                  }
               fname = str_install(sbuf);
               rel_sbuf(sbuf);
               advance_tok(&t1);
               }
            if (t1->tok_id != PpDirEnd)
               errt1(t1, "syntax error in #line");
            if (fname != NULL)
               cs->fname = fname;
            cs->line_adj = n - cs->line_buf[next_char - first_char + 1];
            if (*next_char == '\n')
               ++cs->line_adj;   /* the next lines contains no characters */
            t = next_tok();
            }
         break;

      case MacExpand:
         t = mac_tok();
         break;
         
      case TokLst:
         tlst = src_stack->u.tlst;
         if (tlst == NULL)
            t = NULL;
         else {
            t = copy_t(tlst->t);
            src_stack->u.tlst = tlst->next;
            }
         break;

      case PasteLsts:
         return paste();
      }

   if (t == NULL) {
      pop_src();
      return next_tok();
      }
   else
      return t;
   }

/*
 * Get the next raw white space token, freeing token that the argument
 *  used to point to.
 */
novalue nxt_non_wh(tp)
struct token **tp;
   {
   register struct token *t;

   t = next_tok();
   while (t != NULL && t->tok_id == WhiteSpace) {
      free_t(t);
      t = next_tok();
      }
   free_t(*tp);
   *tp = t;
   }

/*
 * advance_tok - skip past white space after expanding macros and
 *  executing preprocessor directives. This routine may only be
 *  called from within a preprocessor directive because it assumes
 *  it will not see EOF (intput() insures that the terminating
 *  new-line, and thus the PpDirEnd token, will be seen before EOF).
 */
struct token *advance_tok(tp)
struct token **tp;
   {
   struct token *t;

   t = interp_dir();
   while (t->tok_id == WhiteSpace) {
      free_t(t);
      t = interp_dir();
      }
   free_t(*tp);
   *tp = t;
   }

/*
 * merge_whsp - merge a sequence of white space tokens into one token,
 *  returning it along with the next token. Whether these are raw or
 *  processed tokens depends on the token source function, t_src.
 */
novalue merge_whsp(whsp, next_t, t_src)
struct token **whsp;
struct token **next_t;
struct token *(*t_src)();
   {
   struct token *t1;
   struct str_buf *sbuf;
   int line = -1;
   char *fname = "";
   char *s;

   free_t(*whsp);
   t1 = (*t_src)();
   if (t1 == NULL || t1->tok_id != WhiteSpace)
      *whsp  = NULL;
   else {
      *whsp = t1;
      t1 = (*t_src)();
      if (t1 != NULL && t1->tok_id == WhiteSpace) {
         if (whsp_image == NoSpelling) {
            while (t1 != NULL && t1->tok_id == WhiteSpace) {
               free_t(t1);
               t1 = (*t_src)();
               }
            }
         else {
            /*
             * must actually merge white space.
             */
            sbuf = get_sbuf();
            if ((*whsp)->flag & LineChk) {
               line = (*whsp)->line;
               fname = (*whsp)->fname;
               }
            for (s = (*whsp)->image; *s != '\0'; ++s) {
               AppChar(*sbuf, *s);
               if (*s == '\n' && line != -1)
                  ++line;
               }
            while (t1 != NULL && t1->tok_id == WhiteSpace) {
               if (t1->flag & LineChk) {
                  line = t1->line;
                  fname = t1->fname;
                  }
               for (s = t1->image; *s != '\0'; ++s) {
                  AppChar(*sbuf, *s);
                  if (*s == '\n' && line != -1)
                     ++line;
                  }
               free_t(t1);
               t1 = (*t_src)();
               }
            (*whsp)->image = str_install(sbuf);
            rel_sbuf(sbuf);
            if (t1 != NULL && !(t1->flag & LineChk) && line != -1) {
               t1->flag |= LineChk;
               t1->line = line;
               t1->fname = fname;
               }
            }
         }
      }
   *next_t = t1;
   }
