/*
 * This file does most of the memory management. For fixed length blocks,
 *  it maintains free lists. When a list is empty, it mallocs a new block.
 */
#include "preproc.h"

struct src *src_stack = NULL;

static struct macro      *macro_free     = NULL;
static struct token      *token_free     = NULL;
static struct tok_lst    *tok_lst_free   = NULL;
static struct id_lst     *id_lst_free    = NULL;
static struct src        *src_free       = NULL;
static struct char_src   *char_src_free  = NULL;
static struct mac_expand *mac_expand_free = NULL;
static struct paste_lsts *paste_lsts_free = NULL;
static struct str_buf    *str_buf_free    = NULL;

struct macro *new_macro(mname, category, multi_line, prmlst, body)
char *mname;
int category;
int multi_line;
struct id_lst *prmlst;
struct tok_lst *body;
   {
   struct macro *mp;

   if ((mp = macro_free) == NULL)
      NewStruct(mp, macro)
   else
      macro_free = macro_free->next;
   mp->mname = mname;
   mp->category = category;
   mp->multi_line = multi_line;
   mp->prmlst = prmlst;
   mp->body = body;
   mp->ref_cnt = 1;
   mp->recurse = 0;
   mp->next = NULL;
   return mp;
   }

struct token *new_token(id, image, fname, line)
int id;
char *image;
char *fname;
int line;
   {
   struct token *t;

   if ((t = token_free) == NULL)
      NewStruct(t, token)
   else
      token_free = token_free->next;
   t->tok_id = id;
   t->image = image;
   t->fname = fname;
   t->line = line;
   t->flag = 0;
   return t;
   }

struct token *copy_t(t)
struct token *t;
   {
   struct token *t1;

   if (t == NULL)
      return NULL;

   if ((t1 = token_free) == NULL)
      NewStruct(t1, token)
   else
      token_free = token_free->next;
   *t1 = *t;
   return t1;
   }

struct tok_lst *new_t_lst(tok)
struct token *tok;
   {
   struct tok_lst *tlst;

   if ((tlst = tok_lst_free) == NULL)
      NewStruct(tlst, tok_lst)
   else
      tok_lst_free = tok_lst_free->next;
   tlst->t = tok;
   tlst->next = NULL;
   return tlst;
   }

struct id_lst *new_id_lst(id)
char *id;
   {
   struct id_lst *ilst;

   if ((ilst = id_lst_free) == NULL)
      NewStruct(ilst, id_lst)
   else
      id_lst_free = id_lst_free->next;

   ilst->id = id;
   ilst->next = NULL;
   return ilst;
   }

struct char_src *new_cs(fname, f, bufsize)
char *fname;
FILE *f;
int bufsize;
   {
   struct char_src *cs;
   int i;
   int c;
   int line;

   if ((cs = char_src_free) == NULL) {
      NewStruct(cs, char_src)
      cs->bufsize = 0;
      }
   else
      char_src_free = char_src_free->next;

   if (cs->bufsize < bufsize) {
      if (cs->bufsize > 0) {
         free(cs->char_buf);
         free(cs->line_buf);
         }
      if ((cs->char_buf = (int *)malloc(sizeof(int) * bufsize)) == NULL)
         err1("out of memory");
      if ((cs->line_buf = (int *)malloc(sizeof(int) * bufsize)) == NULL)
         err1("out of memory");
      cs->bufsize = bufsize;
      }

   cs->fname = fname;
   cs->f = f;
   cs->line_adj = 0;
   cs->tok_sav = NULL;
   cs->dir_state = CanStart;

   return cs;
   }

struct mac_expand *new_me(m, args, exp_args)
struct macro *m;
struct tok_lst **args;
struct tok_lst **exp_args;
   {
   struct mac_expand *me;

   if ((me = mac_expand_free) == NULL)
      NewStruct(me, mac_expand)
   else
      mac_expand_free = mac_expand_free->next;

   me->m = m;
   me->args = args;
   me->exp_args = exp_args;
   me->rest_bdy = m->body;
   return me;
   }

struct paste_lsts *new_plsts(trigger, tlst, plst)
struct token *trigger;
struct tok_lst *tlst;
struct paste_lsts *plst;
   {
   struct paste_lsts *plsts;

   if ((plsts = paste_lsts_free) == NULL)
      NewStruct(plsts, paste_lsts)
   else
      paste_lsts_free = paste_lsts_free->next;

   plsts->trigger = trigger;
   plsts->tlst = tlst;
   plsts->next = plst;
   return plsts;
   }

struct str_buf *get_sbuf()
   {
   struct str_buf *sbuf;

   if ((sbuf = str_buf_free) == NULL) {
      NewStruct(sbuf, str_buf)
      init_sbuf(sbuf);
      }
   else
      str_buf_free = str_buf_free->next;

   return sbuf;
   }

novalue push_src(flag, ref)
int flag;
union src_ref *ref;
   {
   struct src *sp;

   if ((sp = src_free) == NULL) 
      NewStruct(sp, src)
   else
      src_free = src_free->next;

   sp->flag = flag;
   sp->cond = NULL;
   sp->u = *ref;
   sp->ntoks = 0;

   if (src_stack->flag == CharSrc)
      src_stack->u.cs->next_char = next_char;
   sp->next = src_stack;
   src_stack = sp;
   }
   
novalue free_t(t)
struct token *t;
   {
   if (t != NULL) {
      t->next = token_free;
      token_free = t;
      }
   }

novalue free_t_lst(tlst)
struct tok_lst *tlst;
   {
   if (tlst == NULL)
      return;
   free_t(tlst->t);
   free_t_lst(tlst->next);
   tlst->next = tok_lst_free;
   tok_lst_free = tlst;
   }
   
novalue free_id_lst(ilst)
struct id_lst *ilst;
   {
   if (ilst == NULL)
       return;
   free_id_lst(ilst->next);
   ilst->next = id_lst_free;
   id_lst_free = ilst;
   }
   
novalue free_m(m)
struct macro *m;
   {
   if (--m->ref_cnt != 0)
      return;
   free_id_lst(m->prmlst);
   free_t_lst(m->body);
   m->next = macro_free;
   macro_free = m;
   }

novalue free_m_lst(m)
struct macro *m;
   {
   if (m == NULL)
      return;
   free_m_lst(m->next);
   free_m(m);
   }

novalue free_plsts(plsts)
struct paste_lsts *plsts;
   {
   plsts->next = paste_lsts_free;
   paste_lsts_free = plsts;
   }

novalue rel_sbuf(sbuf)
struct str_buf *sbuf;
   {
   sbuf->next = str_buf_free;
   str_buf_free = sbuf;
   }

novalue pop_src()
   {
   struct src *sp;
   struct char_src *cs;
   struct mac_expand *me;
   int i;

   if (src_stack->flag == DummySrc)
      return;

   sp = src_stack;

   src_stack = sp->next;
   if (src_stack->flag == CharSrc) {
      first_char = src_stack->u.cs->char_buf;
      next_char = src_stack->u.cs->next_char;
      last_char = src_stack->u.cs->last_char;
      }

   if (sp->cond != NULL)
      errt2(sp->cond->t, "no matching #endif for #", sp->cond->t->image);
   switch (sp->flag) {
      case CharSrc:
         cs = sp->u.cs;
         if (cs->f != NULL)
            fclose(cs->f);
         cs->next = char_src_free;
         char_src_free = cs;
         break;
      case MacExpand:
         me = sp->u.me;
         if (me->args != NULL) {
            for (i = 0; i < me->m->category; i++) {
               free_t_lst(me->args[i]);
               free_t_lst(me->exp_args[i]);
               }
            free(me->args);
            free(me->exp_args);
            }
         --me->m->recurse;
         free_m(me->m);
         me->next = mac_expand_free;
         mac_expand_free = me;
         break;
      }
   sp->next = src_free;
   src_free = sp;
   }
