/*
 * The functions in this file maintain a hash table of strings and manage
 *   string buffers.
 */
#include "preproc.h"

/*
 * Prototype for static function.
 */
hidden int s_hash Params((char *s));

struct str_entry {
   char *s;
   struct str_entry *next;
   };

static struct str_entry *str_tbl[HashSize];

/*
 * s_hash - create a hash number based on the first 10 characters of a string.
 */ 
static int s_hash(s)
char *s;
   {
   register int i;
   register unsigned int h;

   h = 0;
   for (i = 1; *s != '\0' && i < 10; ++i)
       h += *s++;
   return h % HashSize;
   }

/*
 * init_str - initialize string hash table.
 */
novalue init_str()
   {
   static int first_time = 1;
   struct rsrvd_wrd *rw;
   struct str_entry *se;
   char *s;
   int h;

   if (first_time)
      first_time = 0;
   else
      return;

   for (h = 0; h < HashSize; ++h)
      str_tbl[h] = NULL;
   }

/*
 * init_sbuf - initialize a new sbuf struct, allocating an initial buffer.
 */
novalue init_sbuf(sbuf)
struct str_buf *sbuf;
   {
   sbuf->size = SBufSize;
   sbuf->strtimage = (char *)malloc(SBufSize);
   sbuf->endimage = sbuf->strtimage;
   sbuf->end = sbuf->strtimage + SBufSize;
   }

/*
 * new_sbuf - allocate a new buffer for a sbuf struct, copying the partially
 *   created string from the full buffer to the new one.
 */
novalue new_sbuf(sbuf)
struct str_buf *sbuf;
   {
   char *s1, *s2;

   sbuf->size *= 2;
   s1 = sbuf->strtimage;
   sbuf->strtimage = (char *)malloc(sbuf->size);
   s2 = sbuf->strtimage;
   while (s1 < sbuf->endimage)
      *s2++ = *s1++;
   sbuf->endimage = s2;
   sbuf->end = sbuf->strtimage + sbuf->size;
   }

/*
 * spec_str - install a special string in the string table.
 */
char *spec_str(s)
char *s;
   {
   struct str_entry *se;
   int h;

   h = s_hash(s);
   for (se = str_tbl[h]; se != NULL; se = se->next) 
      if (strcmp(s, se->s) == 0) {
         return se->s;
         }
   NewStruct(se, str_entry)
   se->s = s;
   se->next = str_tbl[h];
   str_tbl[h] = se;
   return s;
   }

/*
 * str_install - find out if the string in the buffer is in the string table.
 *   If not, put it there. Return a pointer to the string in the table.
 */
char *str_install(sbuf)
struct str_buf *sbuf;
   {
   int h;
   struct str_entry *se;

   AppChar(*sbuf, '\0')
   h = s_hash(sbuf->strtimage);
   for (se = str_tbl[h]; se != NULL; se = se->next) 
      if (strcmp(sbuf->strtimage, se->s) == 0) {
         sbuf->endimage = sbuf->strtimage;
         return se->s;
         }
   NewStruct(se, str_entry)
   se->s = sbuf->strtimage;
   sbuf->strtimage = sbuf->endimage;
   se->next = str_tbl[h];
   str_tbl[h] = se;
   return se->s;
   }
